#####################################################################
# This code is written and copyrighted by Clif Flynt, 2017
#
# It is licenced for public use with these terms:
#
#  This licence agreement must remain intact with the code.
# 
#  You may use this code freely for commercial or non-commercial purposes.
#
#  You may examine and modify the code to your heart's content.
#
#  You may not claim to have authored the code (aside from modifications
#  you may have made.)
#
#  You may not distribute the code without this license agreement
#
#  Maintenance, extension, modification, etc can be performed by:
#
#   Noumena Corporation
#   8888 Black Pine Ln
#   Whitmore Lake, MI  48189
#
#   Contact: clif@noucorp.com
#
#####################################################################

# Self contained info in comments after field definition
#  --- Start of display info comments
#  No --- = No display
#  Tcl list
#  {help} {label} {reference/blank} {entry/combobox args} {?row column?} {?grid arg?}
#
#  {help} Popup help on field
#
#  {label} Label to display with field
#
#  {reference} -- If the line includes "references", then this is a list
#    of fields from the referenced table to use to ID that entry.
#   ie : 
#   personID integer references person --- {...} {...} {person.first person.last} {...}
#   If there is no referenced on this line, this is a list of tables
#   that might be referenced by this table and references is a list of tables
#   that should use this field in a reduced GUI.
#  ie:
#   CREATE TABLE job {
#   personID integer references person --- {...} {...} {person.first person.last} {...}
#   }
#   CREATE TABLE email {
#     id integer,
#     personID integer references person,
#     type text, -- Name of this entry
#     addr text  -- Email addr
#   }
#   CREATE TABLE person {
#     jobID reference job, 
#     payScaleID reference payScale, 
#     work_email,        --- {...} {...} {email.addr}
#     home_email,        --- {...} {...} {email.addr}
#     first_name string, --- {...} {...} {job payscale}
#     last_name string, --- {...} {...} {job payscale}
#     address string
#   }
#  When displaying JOB, the person.first_name and person.last name
#  fields will be used to identify persons associated with that job.
#
# {entry/combobox args} - appended to entry/combobox widgets
#
# {row column} - Optional row and column for display.  Else, builds left-to-right
#
# {grid opts} - Optional args to grid - ie: {columnspan 2} 
#
#  New/Delete/Replace/Close buttons
#
#  WHEN USING DICTS: fields are:
#  fld 		name of field
#  help 	popup help
#  label 	label to associate with this field
#  args 	special args for entry/combobox widget
#  rowcol 	list of row/column
#  gridOpts 	sepcial grid options
#  ref 		table.field to use for combbox entries

#  FROM entries are for faking a schema in which some fields
#    aren't in *this* table, but are referenced from another table
#    and need to look local 
#  fromTbl 	table that references this table (as job references person)
#  fromFld	field to be updated when this is stored.
#  fromFixed	fieldName/Fixed value to be used with this update.
#
#  fromTbl email fromFld addr fromFixed {type work}
#   Would generate an 
#    (insert into email (type, addr, personID) 
#       VALUES 'work', 'a@b', $newID)

package provide editTablesOO 1.0

if {[info exists argv] && ([string first "-testET" $argv] >= 0)} {
  catch {::tcl::tm::path add [pwd]}
}

package require classSupport

oo::class create editTable {

##
## Basic object maintenance methods
##

################################################################
#   classmethod unknown {w args}--
#    Override the unknown message to determine if this is a 
#    megawidget create command
# Arguments
#   w - Name of a widget.  If it doesn't start with . , this is bogus.
#   args - everything else
# 
# Results
#   If w is a window name, a new member of this class is created.
# 
  classmethod unknown {w args} {
    if {[string match .* $w]} {
      [self] new $w {*}$args
      return $w
    }
    next $w {*}$args
  }
  
################################################################
#   destructor--
#    Destroy the object and all clones, referenced objs, 
#    close dBs and remove state variables.
# Arguments
#   NONE
# 
# Results
#   Object and much state associated with it is deleted.
# 
  destructor {
    variable State
    classvar openDBCount

    foreach cl $State(clones) {
      $cl destroy
    }

    foreach no $State(refObjs) {
      # It might already be gone
      catch {$no destroy}
    }
    if {[incr openDBCount -1] == 0} {
      my closeDB
    }
    destroy $State(top)
  }

################################################################
#   constructor {top mixin dbEngine args}--
#    Create a class, initialize it
# Arguments
#   top		Name of toplevel window
#   mixin	Name of class to mix into this
#   dbEngine	Type of DB engine - mySQL, tdbc, sqlite3, etc.
#   args	Args required to initialize connection to DB
# Results
#   New object created, new State table, new connection to DB
# 
  constructor {top mixin dbEngine args} {
    variable State
    classvar openDBCount
    incr openDBCount

    array set State [list \
      top $top            \
      mixin $mixin        \
      dbEngine $dbEngine  \
      args $args          \
    ]
    array set State {
      table {}
      rows {}
      refObjs {}
      refFields {}
      refSel {}
      tableNameDict {}
      mapTables {}
      style labelframe
      dbArgs {}
      clones {}
    }
    
    foreach {k v} $args {
       set k2 [string trimleft $k -]
       if {[info exists State($k2)]} {
         set State($k2) $v
       } else {
         error "$k not valid.\nMust be one of [lsort [array names State]]" \
	       "$k not valid.\nMust be one of [lsort [array names State]]"
       }
    }

    set fr2 [info level [expr [info level]-2]]

    # Check for "" mixin for standalone test.
    
    if {$mixin ne ""} {
      oo::objdefine [self] mixin $mixin
      if {[lsearch $fr2 clone] < 0} {
        my init $dbEngine {*}$State(dbArgs)
      }
    }

    set top [frame $top -class Edittable]

    rename ::$top ::$top,frame
    rename [self] ::$top
  }

################################################################
#   method clone {top }--
#    Clone the current db connection and make a new GUI
# Arguments
#   top		New widget name
# 
# Results
#   Creates a new object with same db and settings as existing
#   object.  New object can hold new tables, reference, etc.
# 
#   This is primarily to support displaying references *to* a table
#   as well as references from a table.

  method clone {top } {
    variable State
    [info object class [self]] $top $State(mixin) $State(dbEngine) {*}$State(args)

    foreach fld {db tableNameDict mapTables} {
      catch {$top config -$fld $State($fld)}
    }
    lappend State(clones) $top
    return $top
  }
  
##
## Holders for methods implemented in mixins
##  
################################################################
#   method getTables {} --
#    Holder to let self-tests work.  Not invoked in real life
# Arguments
#   NONE
# 
# Results
#   empty string
# 
  method getTables {} {return ""}
  
################################################################
#   method parray {}--
#    utility function for debugging, prints State table
# Arguments
#   NONE
# 
# Results
#   Generates output to stdout
# 
  method parray {} {
    variable State
    parray State
  }

################################################################
#   method getPrimary {tbl}--
#    Holder to let self-tests work.  Not invoked in real life
# Arguments
#   tbl		Name of the table
# 
# Results
#   None.
# 
  method getPrimary {tbl} {
    return {}
  }
  
##
## General purpose support methods
## 
 
################################################################
#   method removeCurlies {}--
#    Remove curly braces from outside of value fields in State
# Arguments
#   NONE
# 
# Results
#   Modifies all value,* elements in State array.
# 
  method removeCurlies {} {
    variable State
    foreach in [array names State value,*] {
      set State($in) [string trim $State($in) "{}"]
    }
  }


################################################################
#   method getDict {fld}--
#    Return the dict that has a key/value pattern in it
# Arguments
#   ptn		Key for dict element
# 
# Results
#   No side effects
# 
  method getDict {fld} {
    variable State
    return [dict get $State(dicts) $fld]
  }  

################################################################
#   method setDict {ptn}--
#    Replace value of dict with new val
# Arguments
#   ptn		A pattern to search for in the list of dicts
#   dct		New value for dictionary
# 
# Results
#   No side effects
# 
  method setDict {fld dct} {
    variable State
    dict set State(dicts) $fld $dct
  }  


################################################################
#   method getIDFromCombo {fld}--
#    Return the ID for the value selected in a combobox used 
#    for a specific field
# Arguments
#   fld		The name of field referenced by combo box.
# 
# Results
#   No side effects
# 
  method getIDFromCombo {fld} {
    variable State
    
    if {[info exists State(ref,$fld)] } {
      # If there is a ref field, this is a combobox associated with a 
      # reference table
      if {[dict exists $State(idDict,$State(ref,$fld)) [$State(combo,$fld) get]]} {
        return [dict get $State(idDict,$State(ref,$fld)) [$State(combo,$fld) get]]
      } else {
        return ""
      }
    } else {
      # If no ref,* field, this is a combobox with n values determined by
      # a 'values' parameter to the makeGUIwDict call.
      if {[info exists State(value,$fld)]} {
        return $State(value,$fld)
      } else {
        return ""
      }
    }
  }

################################################################
#   method configure {key {val NoSuChVaL}}--
#   method config {key {val NoSuChVaL}}--
#    Modify or report State indices 
# Arguments
#   key		The name of a State array index, may be -key
#   val		Optional new value to assign to this index.  If not
#		present, the current value is returned.
# 
# Results
#   State table may be modified.
# 
  method configure {key {val NoSuChVaL}} {
    my config $key $val
  }

  method config {{key ""} {val NoSuChVaL}} {
    variable State
    if {$key eq ""} {
       set rtn ""
       foreach in [lsort -dict [array names State]] {
         lappend rtn $in $State($in)
       }
       return $rtn
    } elseif {$val eq "NoSuChVaL"} {
      set key [string trimleft $key -]
      return $State($key)
    } else {
      set key [string trimleft $key -]
      set State($key) $val
      return $State($key)
    }
  }

################################################################
#   method clearValues {}--
#    Clear all value indices in State array
# Arguments
#   NONE
# 
# Results
#   all value,* fields are deleted.  The GUI entry and Selection
#   widgets will become blank.
# 
  method clearValues {} {
    variable State
    foreach in [array names State value,*] {
      set State($in) ""
    }
  }

################################################################
#   method clearState {}--
#    Clean up the State variable for a new table/GUI
# Arguments
#   NONE
# 
# Results
#   State array is wiped, objs are destroyed
# 
  method clearState {} {
    variable State

    catch {destroy {*}[winfo children $State(top)]}
    set wipeList [list label field value combo]
    foreach w $wipeList {
      foreach in [array names State $w,*] {
	  unset State($in)
      }
    }
  }

################################################################
#   method reduceSchema {schema}--
  # strip a schema down to the data definitions
# Arguments
#   schema
# 
# Results
#   no side effects.
# 
  method reduceSchema {schema} {
    # If it's already stripped (no create table), just return it.
    if {[string first "create" [string trim [string tolower $schema] " 	{}"]] != 0} {
      return $schema
    }
    set p1 [string first "(" $schema]
    if {$p1 < 0} {return $schema}
    set p2  [string last ")" $schema]
    incr p1
    incr p2 -1
    # puts "TO RETURN: [string range $schema $p1 $p2]"
    return [string range $schema $p1 $p2]
  }
  
################################################################
#   method splitSchemaLine {line}--
#    Splits a line from a schema into the SQL portion
#    name text, -- NAME
#    and screen definition portion
#    --- AA BB CC
# Arguments
#   line	A line from the schema
# 
# Results
#   NO side effects
# 
  method splitSchemaLine {line} {
     set p1 [string first , $line]
     if {$p1 < 0} {
       set p1 [string first -- $line]
       if {$p1 < 0} {
         set p1 [string length $line]
       }
     }
     set def [string trim [string range $line 0 $p1]]
     set p2 [string first "---" $line]
     if {$p2 > 0} {
       set info [string trim [string range $line $p2 end] "- "]
     } else {
       set info ""
     }
     return [list $def $info]
   }

################################################################
#   method getSchemaLine {schema element}--
#    extract a line from a schema based on a field name
# Arguments
#   schema	A newline delimited schema, one field/line
#   element	Name of a field
#
# Results
#   No side effects
# 
  method getSchemaLine {schema element} {
    set element [string tolower $element]
    foreach l [split $schema \n] {
      set l [string tolower [string trim $l]]
      if {[string first $element $l] == 0} {
        return $l
      }
    }
    return ""
  }

################################################################
#   method getSchemaLineRef {schema table}--
#    Return line from schema that references a table
# Arguments
#   schema	A newline delimited schema, one field/line
#   table	Name of a table being referenced.
# 
# Results
#   NO side effects.
# 
  method getSchemaLineRef {schema table} {
    set table [string tolower $table]
    foreach l [split $schema \n] {
      set l [string tolower [string trim $l]]
      if {([string first " $table" $l] >= 0) && \
          ([string first "references" $l] > 0)} {
        return $l
      }
    }
    return ""
  }
  
################################################################
#   method cleanSchemaLine {line}--
#    Remove spaces, comments
# Arguments
#   line 	A line from a schema
# 
# Results
#   Same line, trimmed of spaces commas, () {} and comments
# 
  method cleanSchemaLine {line} {

    if {[set p [string first "--" $line]] > 0} {
      set line [string range $line 0 $p]
    }
    set line [string trim [string tolower $line]]
    set line [string trim $line " {} (),-"]
    return $line
  }
  
################################################################
#   method getRefsForFieldFromSchema {schema targetTbl}--
#    Extract list of references from the --- portion of schema
# Arguments
#   schema	A newline delimited schema, one field/line
#   targetTbl	Table to find references to
# 
# Results
#   
# 
  method getRefsForFieldFromSchema {schema targetTbl} {

    set lst [split $schema \n]
    set p [lsearch $lst "*references*${targetTbl}*"]
    set line [lindex $lst $p]

    lassign [my splitSchemaLine $line] def info
    lassign $info help label refs args rowcol gridOpts

    foreach ref $refs {
      lassign [split $ref .] tbl fld
      lappend flds $fld
    }
    return [list $tbl $flds]
  }

################################################################
#   method getReferencingTables {tbl}--
#    Retrieve list of tables that reference this table
# Arguments
#   tbl		Name of current table
# 
# Results
#   No side effects
# 
  method getReferencingTables {tbl} {
    variable State
    set refs ""
    
    set tbl [string tolower $tbl]

    foreach ref [my getTables] {
      set sch [my getSchema $ref 0]
      if {[string first $tbl $sch] > 0} {
        foreach l [split $sch \n] {
          lassign [my splitSchemaLine [string tolower $l]] def info
	  if {[string match "*references *$tbl\[ ,\]*" $def] &&
	    ([string first $tbl $info] > 0)} {
# puts "REF: $ref L: $l"
	    lappend refs $ref
	  }
	}
      }
    }
    return $refs
  }

##
## GUI Creation methods
##
################################################################
#   method makeGUI {schema {style full} }--
#    GUI for editing a table
# Arguments
#   schema	A schema to generate GUI for
#   style	Whether to be a full GUI with all references
#		or a reduced GUI with only the immediate fields.
# Results
#   Windows are created, State is updated
# 
  method OLDmakeGUI {schema {style full} } {
    variable State
# puts "    makeGUI schema $schema \n style $style "
    set top $State(top)
    
    foreach obj $State(refObjs) {
      $obj destroy
    }
    set State(refObjs) {}
    set State(refFields) {}
    set State(refSel) {}

    set row 1
    set col 1
    set uniq 0
    set maxRow 0
    
    if {[string first "---" $schema] < 0} {
      set instrumented 0
    } else {
      set instrumented 1
    }

    foreach l [split [my reduceSchema $schema] \n] {
      set l [string trim $l]
      if {[string first "unique" $l] > 0} {continue}
      set fld [string range $l 0 [string first " " $l]]
      set fld [string trim $fld]
      
      switch $instrumented {
        0 {
	  set label $fld
	  set ref ""
	  set args ""
	  set rowcol ""
	  set gridOpts ""
	  set help "No Help for $fld"
	}
        1 {
          if {[set p [string first "---" $l]] < 0} {continue}
          set l2 [string range $l $p end]
          set rowcol ""
          set ref ""
# puts "L2: $l2"
          lassign $l2 d help label ref args rowcol gridOpts
          if {$rowcol ne ""} {
            lassign $rowcol row col
            if {$row > $maxRow} {
    	      set maxRow $row
    	    }
          }
	}
      }
      
      if {$label eq ""} {continue}
      set State(label,$fld) $label
      set State(field,$label) $fld

      set f [labelframe $top.[incr uniq] -text $label]
      if {[string first references $l] < 0} {
        set w [entry $f.e -textvar [my varname State(value,$fld)] {*}$args]
	if {[info exists State(validate,$fld)]} {
# puts "SET VALIDATE - $w $fld"
	  bind $w <FocusOut> [list {*}$State(validate,$fld) $w $fld]
	}
      } else {
        set w [ttk::combobox $f.e -textvariable \
	    [my varname State(value,$fld)] {*}$args]
# source debug.tcl
# trace add variable [my varname State(value,$fld)] write traceVar
	set State(combo,$fld) $w
	set s1 [my getValues $ref]
        set State(ref,$fld) [lindex $ref end]
        set State(refs,$fld) $ref
	$w configure -values [lsort -dict $s1]
      }
      grid $w -sticky news -row 0 -column 0
      grid rowconfigure $f 0 -weight 1
      grid columnconfigure $f 0 -weight 1
      grid $f -row $row -column $col -pady 2 -sticky news {*}$gridOpts
      grid rowconfigure $top $row -weight 1
      grid columnconfigure $top $col -weight 1

      if {$help ne ""} {
        balloon $w $help
      }
      
      incr col
      if {$col > 5} {
        set col 1
	incr row
      }

    }
    set row $maxRow
    incr row 2

    if {[info exists State(table)]} {
      set row [my makeMyRefObjs $State(table) $row]
    }
    
    set f [frame $top.buttons]
    set State(buttonFrame) $f
    grid $f -row $row -column 0 -columnspan 5
    
    set texts {New Delete Replace Search } 
    set helps { {Create a new row in database} 
	          {Delete database row for this data}
		  {Replace the original data with new values}
		  {Search for records} 
		  }

    if {[string tolower $style] eq "full"} {
      lappend texts Close 
      lappend helps {Abort editing this row}
    }
    
    foreach txt $texts  hlp $helps {
      set w [button $f.b_$txt -text $txt -command "[self] do$txt"]
      pack $w -side left
      balloon $w $hlp
    }

  }

################################################################
#   method makeGUI {schema {style full} }--
#    GUI for editing a table
# Arguments
#   schema	A schema to generate GUI for
#   style	Whether to be a full GUI with all references
#		or a reduced GUI with only the immediate fields.
# Results
#   Windows are created, State is updated
# 
  method makeGUI {schema {style full} } {
    variable State
 # puts "    makeGUI schema $schema \n style $style "
    set top $State(top)
    
    foreach obj $State(refObjs) {
      $obj destroy
    }
    set State(refObjs) {}
    set State(refFields) {}
    set State(refSel) {}

    set row 1
    set col 1
    set uniq 0
    set maxRow 0
    
    set dicts [my genDictFromSchema $schema]
    my makeGUIwDict $dicts $style

  }

  method makeGUIwDict {dicts {style full} } {
    variable State

#  puts "    makeGUIwDict dict $dicts \n style $style "
#  dict for {k1 v1} $dicts {
#    dict for {k v} $v1 {
#     puts "$k1 - $k: $v"
#    }
#  }
    set top $State(top)
    
    foreach obj $State(refObjs) {
      $obj destroy
    }
    set State(refObjs) {}
    set State(refFields) {}
    set State(refSel) {}

    set row 1
    set col 1
    set uniq 0
    set maxRow 0
    
    dict for {fld def} $dicts {
      set gridOpts {}
      set fld $fld
      set ref ""
# puts "DEF: $def"
   # Do not display primary unique fields, but do set the State index.
      if {[dict exists $def primary] || [dict exists $def unique]} {
        set State(primary) $fld
	continue
      }

      dict for {k v} $def {
# puts "        set $k $v"
        set $k $v
      }
      
      if {[info exists primary]} {
# puts "        set State(primary) $primary"
        set State(primary) $primary
      }
      
      set State(label,$fld) $label
      set State(field,$label) $fld

      set f [labelframe $top.[incr uniq] -text $label]

      if {$ref eq ""} {
        if {(![info exists values]) || ($values eq "")} {
          set w [entry $f.e -textvar [my varname State(value,$fld)] {*}$args]
        } else {
          set w [ttk::combobox $f.e -textvariable \
	      [my varname State(value,$fld)] -values $values {*}$args]
	  set State(combo,$fld) $w
	}
	if {[info exists State(validate,$fld)]} {
	  bind $w <FocusOut> [list {*}$State(validate,$fld) $w $fld]
	}
      } else {
        set w [ttk::combobox $f.e -textvariable \
	    [my varname State(value,$fld)] {*}$args]
	set State(combo,$fld) $w
	set s1 [my getValues $ref]
        set State(ref,$fld) [lindex $ref end]
        set State(refs,$fld) $ref
	$w configure -values [lsort -dict $s1]
      }
      grid $w -sticky news -row 0 -column 0
      grid rowconfigure $f 0 -weight 1
      grid columnconfigure $f 0 -weight 1
      if {[info exists rowcol] && ([string trim $rowcol] ne "")} {
        lassign $rowcol row col
	if {$row > $maxRow} {
	  set maxRow $row
	}
      }
      grid $f -row $row -column $col -pady 2 -sticky news {*}$gridOpts
      grid rowconfigure $top $row -weight 1
      grid columnconfigure $top $col -weight 1

      if {$help ne ""} {
        balloon $w $help
      }
      
      incr col
      if {$col > 5} {
        set col 1
	incr row
	if {$row > $maxRow} {
	  set maxRow $row
	}
      }

      foreach {k v} $def {
        catch {unset $k}
      }
    }

    set row $maxRow
    incr row 2

    if {[info exists State(table)]} {
      set row [my makeMyRefObjs $State(table) $row]
    }
    
    set f [frame $top.buttons]
    set State(buttonFrame) $f
    grid $f -row $row -column 0 -columnspan 5
    
    set texts {New Delete Replace Search } 
    set helps { {Create a new row in database} 
	          {Delete database row for this data}
		  {Replace the original data with new values}
		  {Search for records} 
		  }

    if {[string tolower $style] eq "full"} {
      lappend texts Close 
      lappend helps {Abort editing this row}
    }
    
    foreach txt $texts  hlp $helps {
      set w [button $f.b_$txt -text $txt -command "[self] do$txt"]
      pack $w -side left
      balloon $w $hlp
    }
# puts "DONE WITH MAKEGUI"
  }



################################################################
#   method genDictFromSchema {schema }--
#    Generate a dictionary for building a GUI from the schema
# Arguments
#   schema	A schema from an SQL table.  May be "normal"
#		(no --- definitions) or 
#               instrumented (has --- help label ref args rowcol gridOpts)
# 
# Results
#   No side effects.
# 
  method genDictFromSchema {schema} {
    variable State

    set top $State(top)
    
    foreach obj $State(refObjs) {
      $obj destroy
    }
    set State(refObjs) {}
    set State(refFields) {}
    set State(refSel) {}

    set row 1
    set col 1
    set uniq 0
    set maxRow 0
    
    if {[string first "---" $schema] < 0} {
      set instrumented 0
    } else {
      set instrumented 1
    }
    
    set dicts [list]
    if {[regexp -nocase {.*table create.*} $schema]} {
# puts "SETTING STATE TABLE: $l1"
      set State(table) [lindex $l1 2]
    }

    foreach l [split [my reduceSchema $schema] \n] {

      set fromTbl 	{}
      set fromFld	{}
      set fromFixed	{}
      set values 	{}
      set fixed 	{}

      set l [string trim $l]
      if {[set p [string first --- $l]] > 0} {
        set sl [string range $l 0 $p]
      } else {
        set sl $l 
      }
      # Sudden death tests.
      set ll [string tolower $sl]
      if {([string first "unique" $ll] > 0) || ([string first "primary" $ll] > 0)} {
        set State(primary) [lindex $ll 0]
        continue
      }
      if {[llength [list {*}$l]] < 2} {continue}

      set fld [string range $l 0 [string first " " $l]]
      set fld [string trim $fld]
      
      switch $instrumented {
        0 {
	  if {[string tolower $fld] eq "table"} {continue}
	  if {[string first primary $l] > 1} {continue}
	  set label $fld
	  set ref ""
	  set args ""
	  set rowcol ""
	  set gridOpts ""
	  set help "No Help for $fld"
	}
        1 {
          if {[set p [string first "---" $l]] < 0} {continue}
          set l2 [string range $l $p end]
          set rowcol ""
          set ref ""
          lassign $l2 d help label ref args rowcol gridOpts
          if {$rowcol ne ""} {
            lassign $rowcol row col
            if {$row > $maxRow} {
    	      set maxRow $row
    	    }
          }
	}
      }
# puts "XX [string first reference $l] REF: $ref "
      # This is a "from" reference.
      if {[lsearch $ref VALUES] == 0} {
        set values [lrange $ref 1 end]
	set ref {}
      } elseif {([string first reference $l] < 0) && ($ref ne "")} {
        if {[string first . $ref] > 0} {
	  lassign [split $ref .] ftbl ffld
          set fromTbl $ftbl
	  set fromFld $ffld
        } else {
          set fromTbl $ref
  	  set fromFld {}
        }
	set ref ""
      }
#      lappend dicts [dict create fld $fld help $help label $label ref $ref \
#          args $args rowcol $rowcol gridOpts $gridOpts fixed $fixed \
#	  fromTbl $fromTbl fromFld $fromFld values $values]

       dict append dicts $fld [dict create help $help label $label ref $ref \
          args $args rowcol $rowcol gridOpts $gridOpts fixed $fixed \
	  fromTbl $fromTbl fromFld $fromFld values $values]

    }
    set State(dicts) $dicts
    return $dicts
  }

################################################################
#   method makeRefGUI {schema}--
#    Make a GUI for a table that references another table.
#    This is a reduced info GUI
# Arguments
#   tbl    - The name of the table being referenced.
#   schema - The table schema
# 
# Results
#   $top is populated with new data
# 
  method makeRefGUI {tbl schema parent} {
    variable State
    
    set top $State(top)

    set col 2
    set row 0
    grid rowconfigure $top 0 -weight 1
    
    if {[lsearch $State(mapTables) $State(table)] < 0} {
      set noRef 1
    } else {
      set noRef 0
    }

    foreach l [split $schema \n] {
      if {$noRef && [string first references $l] > 0} {continue}
      set l [string trim $l]
      if {[set p [string first "---" $l]] < 0} {continue}
      set l2 [string range $l $p end]
      lassign $l2 d help label ref args rowcol gridOpts

      if {[lsearch $ref $tbl] >= 0} {
        set fld [string range $l 0 [string first " " $l]]
        set fld [string trim $fld]
        set f [labelframe $top.[incr uniq] -text $label]
        set w [entry $f.e -textvar [my varname State(value,$fld)] {*}$args]
        grid $w -sticky news -row 0 -column 0
        grid rowconfigure $f 0 -weight 1
        grid columnconfigure $f 0 -weight 1
        grid $f -row $row -column $col -pady 3 -sticky news {*}$gridOpts
        grid columnconfigure $top $col -weight 1
        if {$help ne ""} {
          balloon $w $help
        }
        incr col
        if {$col > 5} {
          set col 1
  	  incr row
        }
      } elseif {($noRef == 0) && ([string first . $ref] > 0)} {
          lassign [split [lindex $ref 0] .] name fld
	  if {[string tolower $name] eq [string tolower $tbl]} {
	    continue
	  }
	  foreach rr $ref {
            lassign [split $rr  .] name fld
  	    set ll [my getSchemaLine [my getSchema $name 0] $fld]
	    lassign [my splitSchemaLine $ll] def info
            lassign $info help label refs args rowcol gridOpts
            set f [labelframe $top.[incr uniq] -text $label]
#            set w [entry $f.e -textvar [my varname State(value,$fld)] {*}$args]
            set w [entry $f.e -textvar [my varname State(value,$rr)] {*}$args]
            grid $w -sticky news -row 0 -column 0
            grid rowconfigure $f 0 -weight 1
            grid columnconfigure $f 0 -weight 1
            grid $f -row $row -column $col -pady 3 -sticky news {*}$gridOpts
            grid rowconfigure $top $row -weight 1
            grid columnconfigure $top $col -weight 1
	    incr col
	  }
      }
    }
    
    button $top.mod -text "E" -padx 0 -pady 0 \
        -command "[self] newEdit $State(table) $parent" 
    grid $top.mod -row 0 -column 0
    scale $top.select -command "[self] populateRow" -from 0 \
          -to 1 -length 40\
          -orient vertical
    grid $top.select -row 0 -column 1 -sticky ns
    set State(refSel) $top.select
  }
  
################################################################
#   method newEdit {tbl parent}--
#    Create a new edit window, attached to an existing display
#    The primary purpose for this is to allow a user to easily
#    add new data to a table referenced by the currently displayed
#    table.
#
# Arguments
#   tbl		The new table to display
#   parent	The parent window
# Results
#   New object is created with separate State.
#   DB may be modified by new object.
# 
  method newEdit {tbl parent} {
    variable State

    catch {destroy .edit_$tbl}
    set t [toplevel .edit_$tbl]
    wm title $t "Modify [string totitle $tbl]"

    set no [my clone $t.edt]
    set sch [$no getSchema $tbl]
    $no makeGUI $sch
    set sql {}

    set prim [my getPrimary $State(table)]
    if {[info exists State(value,$prim)]} {
      set sql "$prim = $State(value,$prim)"
    } else {
      foreach nm [array names State value,*] {
        if {([string first "." $nm] < 0) && ($State($nm) ne "")} {
          lassign [split $nm ,] vv fld
	  if {[string is double $State($nm)]} {
	    set val $State($nm)
	  } else {
	    set val "'$State($nm)'"
	  }
          append sql "$fld = $val AND "
        }
      }
      set sql [string trim $sql " AND"]
    }
    if {$sql ne ""} {
      $no populateBySearch $sql
    }
    grid $no -sticky news -row 0 -column 0
    grid rowconfigure $t 0 -weight 1
    grid columnconfigure $t 0 -weight 1
    bind $no <Destroy> "destroy $t; $parent rebuildGUI"
  }
  
################################################################
#   method rebuildGUI {}--
#    clear and rebuild GUI.
#    Primary use for this is rebuilding a GUI after the underlying
#    database has been modified.  This will put new values into
#    comboboxes.
# Arguments
#   NONE
# 
# Results
#   GUI is reworked.  State is modified.  
# 
  method rebuildGUI {} {
    variable State


    destroy {*}[winfo children [string trim [self] :]]
    update idle;
    my makeGUI [my getSchema $State(table)]
    update idle;
    set prim $State(primary)
    if {[info exists State(value,$prim)]} {
      my populateBySearch "$State(primary) = $State(value,$prim)"
    }
  }

################################################################
#   method makeMapGUI {mapTbl pullTbl checkTbl}--
#    Make a GUI for tables that map from tableA to tableB
# Arguments
#   mapTbl	Name of the mapping table
#   pullTbl	tableA name
#   checkTbl	tableB name
# 
# Results
#   May modify DB 
#   NOTE: This proc may need work - can it delete references?
  method makeMapGUI {mapTbl pullTbl checkTbl} {
    variable State
    
    labelframe $State(top).left \
        -text "Select an item from [msgcat::mc $pullTbl]"
    labelframe $State(top).right \
        -text "Select appropriate items from [msgcat::mc $checkTbl]"
    
    
    # Clear out selection vars from potiential previous
    # installs
    foreach in [array names State pull,*] {
      unset State($in)
    }
    foreach in [array names State check,*] {
      unset State($in)
    }

    set mapSchema [my getSchema $mapTbl 0]

    lassign [my getRefsForFieldFromSchema $mapSchema $pullTbl] tbl flds
    set pullPrim [my getPrimary $pullTbl]
    lappend flds $pullPrim

# puts "SELECT [join $flds ,] FROM $tbl"
    set rtn [my doSQL "SELECT [join $flds ,] FROM $tbl"]
    
    foreach $flds $rtn {
      set val {}
      foreach f $flds {
        if {$f eq $pullPrim} {
	  set prim [set $f]
	} else {
          append val "[set $f] "
	}
      }
      set State(pull,$val) $prim
      lappend pullVals $val
    }
    
    set cmbo [ttk::combobox $State(top).left.pull -values $pullVals \
        -textvar [my varname State(pullVal)] ]
    grid $cmbo

    lassign [my getRefsForFieldFromSchema $mapSchema $checkTbl] tbl flds
    set checkPrim [my getPrimary $checkTbl]
    lappend flds $checkPrim
    
    set rtn [my doSQL "SELECT [join $flds ,] FROM $tbl"]
    
    set unique 0
    foreach $flds $rtn {
      set val {}
      foreach f $flds {
        if {$f eq $checkPrim} {
	  set prim [set $f]
	} else {
          append val "[set $f] "
	}
      }
      set bb [checkbutton $State(top).right.b_[incr unique] \
          -text "$val" -variable [my varname State(check,$prim)] \
	  -onvalue 1 -offvalue 0]
      grid $bb -sticky w
    }
    
    
    set bf [frame $State(top).buttons]
    
    foreach txt {{Add Links} {Delete Links}} {
      set w [button $bf.b_[incr unique] -text $txt \
          -command "[self] do_Map[lindex $txt 0] $mapTbl $pullTbl $checkTbl"]
      pack $w -side left	  
    }
    
    grid $State(top).left $State(top).right -sticky ns
    grid $bf -columnspan 2 -row 99
  }
  
################################################################
#   method do_MapAdd {mapTbl pullTbl checkTbl}--
#    Perform the action when a user selects action in MapGUI window
# Arguments
#   mapTbl	Name of map table
#   pullTbl	tableA
#   checkTbl    tableB
# 
# Results
#   Adds new links in a map table
# 
  method do_MapAdd {mapTbl pullTbl checkTbl} {
    variable State
    
# puts "do_MapAdd mapTbl $mapTbl pullTbl $pullTbl checkTbl $checkTbl "
    set mapSchema [my getSchema $mapTbl 0]

    set lst [split $mapSchema \n]
    set p [lsearch $lst "*references*${pullTbl}*"]
    set pullLine [string trim [lindex $lst $p]]
    set p [lsearch $lst "*references*${checkTbl}*"]
    set checkLine [string trim [lindex $lst $p]]
    
    set p1 [string first " " $pullLine]
    set pullFld [string trim [string range $pullLine 0 $p1]]

    set p1 [string first " " $checkLine]
    set checkFld [string trim [string range $checkLine 0 $p1]]
    
    set pullID $State(pull,$State(pullVal))
    foreach in [array names State check,*] {
# puts "IN: $State($in)"
      if {$State($in) != 0} {
        lassign [split $in ,] ch checkID
#  puts "INSERT INTO $mapTbl ($pullFld, $checkFld) VALUES ($pullID, $checkID)"
        my doSQL "INSERT INTO $mapTbl ($pullFld, $checkFld) VALUES ($pullID, $checkID)"
      }
    }
  }

  

################################################################
#   method makeMyRefObjs {tbl row}--
#    Create associated values for tables that reference the current table
# Arguments
#   tbl		The current table being displayed
#   row		Row on which to display the new windows.
# Results
#   GUI is modified, State array is modified.
# 
  method makeMyRefObjs {tbl row} {
    variable State
# puts "makeMyRefObjs tbl $tbl row $row ME: [self]"

    set top $State(top)

    foreach ref [my getReferencingTables $tbl] {
# puts "REF1: $ref"
      set sch [my getSchema $ref 0]
      set schL [split $sch \n]
      set ll [lsearch -inline $schL "*references *$tbl\[ ,\]*"]
      lassign [my splitSchemaLine $ll] def info
      lassign $info h2 l2 r2 e2 r2 g2
      if {[dict exist $State(tableNameDict) $ref]} {
        set show [dict get $State(tableNameDict) $ref]
      } else {
        set show [string totitle $ref]
      }
      set ff [labelframe $top.f_$ref -text "Associated values for $show"]
      grid $ff -row $row -column 0 -sticky news -pady 5 -columnspan 6
      set no [my makeRefObj $ff.eobj $ref $State(table)]
      lappend State(refObjs) $no
      set fld [lindex $ll 0]
      lappend State(refFields) $fld
      grid $no
      incr row
    }
    return $row
  }

  
################################################################
#   method makeRefObj {objName ref tbl}--
#    Make an object that references table $ref from table $tbl
# Arguments
#   objName	Name of new object/window
#   ref		Table being referenced
#   tbl		Current table
# 
# Results
#   
# 
  method makeRefObj {objName ref tbl} {
    variable State
# puts "MakeREFOBJ: ON: $objName REF: $ref TBL: $tbl"
    set no [my clone $objName]
    set sch [$no getSchema $ref]
    $no makeRefGUI $tbl $sch [self]
    if {[lsearch $State(mapTables) $ref] >= 0} {
       $no makeFwdRefObjs $ref $tbl
    }
    return $no
  }
  
################################################################
#   method makeFwdRefObjs {tbl backRef}--
#    make a set of forward reference tables
# Arguments
#   tbl		The table to search
#   backRef	UNUSED ?WHY is it here?
# 
# Results
#   New objects, new windows, Modified State array
# 
  method makeFwdRefObjs {tbl backRef} {
    variable State
# puts "--- makeFwdRefObjs - $tbl $backRef [self]"
    set uniq 0
    set sch [my getSchema $tbl]
    foreach ll [split $sch \n] {
      lassign [my splitSchemaLine $ll] def info
      set def [string tolower $def]
      if {[scan $def "%s references %s" aa bb] == 2} {
# puts "+++ [scan $def "%s references %s" aa bb] $def"
        set bb [string trim $bb " ,"]
        set objNm [string trim [self].fr[incr uniq] :]
# puts "        set no 'my clone $objNm'"
        set no [my clone $objNm]
	$no getSchema $bb
	lappend State(refObjs) $no
	lappend State(refFields) [$no getPrimary $bb]
      }
    }
  }
  
################################################################
#   method populate {fldValDict}--
#    Populate the State array to display values
# Arguments
#   fldValDict	A dict of fields and values
# 
# Results
#   State array value,* indices are modified.
# 
  method populate {fldValDict} {
    variable State
#  puts "POPULATE [self] - $fldValDict"    
# puts "POPULATE [self] - FROM: [info level -1]"
    set prim [my getPrimary $State(table)]

    foreach {fld v} $fldValDict {
#  puts "FLD: $fld V: $v"
      if {[info exists State(combo,$fld)]} {
        if {[info exists State(ref,$fld)]} {
          if {$v ne ""} {
            # This was removed in ROMaN variant
	    # lassign [split $State(ref,$fld) .] tbl ref
    	    set State(value,$fld) ""
	    foreach rr $State(refs,$fld) {
	      # This line was added in ROMaN variant
              lassign [split $rr .] tbl ref
#  puts "COMBO: $State(combo,$fld) - REF: $State(ref,$fld) -- $tbl $ref"
  	      append State(value,$fld) [my getValueByRef $tbl $ref $State(primary,$State(ref,$fld)) $v]
	    }
	  }
	} else {
# puts "SETTING: set State(value,$fld) $v"
	  set State(value,$fld) $v
	}
      } else {
        set State(value,$fld) $v
      }
    }
    if {($State(refObjs) ne "") && [dict exists $fldValDict $prim]} {
      set val [dict get $fldValDict $prim]
      foreach no $State(refObjs) nm $State(refFields) {
        set tt [$no config -table]
	if {[lsearch $State(mapTables) $tt] >= 0} {
	  # puts "MAPSEARCH: NO: $no In: $State(table) TBL: $tt LCLREF: $State(refFields) -- [$no config -refFields] "
	  $no fillFromFwdRef $nm $val
	} else {
          $no clearValues
          $no populateBySearch "$nm = $val"
	}
      }
    }
    foreach in [array names State combo,*] {
      set id [$State($in) current]
      if {$id < 0} {
        lassign [split $in ,] c nm
	if {[info exists State(value,$nm)]} {
	  set pos [lsearch [$State($in) cget -values] "$State(value,$nm)*" ]
# puts "POS: $pos FROM: lsearch [$State($in) cget -values] '$State(value,$nm)*'"
	  if {$pos >= 0} {
	    $State($in) current $pos
	  }
	}
      }
    }
#    foreach in [array names State value,*] {
#      set State($in) [string trim $State($in) "{}"]
#    }
  }
   
################################################################
#   method fillFromFwdRef {key val}--
#    Fill a GUI based on key/value pair
# Arguments
#   key		A field name
#   val		Value to select
# Results
#   GUI is modified, State array is modified
# 
  method fillFromFwdRef {key val} {
    variable State
# puts "fillFromFwdRef key $key val $val - [self]"
# my parray
    my populateWithFwdRef "$key == $val"
  }

#
# Methods invoked by GUI buttons
#
################################################################
#   method doNew {}--
#    Dummy method Actual method is in DB code
# Arguments
#   NONE
# 
# Results
#   NONE
# 
  method doNew {} {
    variable State
  }

################################################################
#   method doDelete {}--
#    Delete a row in the target DB
# Arguments
#   NONE
# 
# Results
#   Modifies DB
# 
  method doDelete {} {
    variable State
    
    set sql "DELETE FROM $State(table) WHERE $State(primary) = $State(value,$State(primary))"
# puts "SQL: $sql"
    my doSQL $sql
    foreach in [array names State value,*] {
      set State($in) ""
    }
  }

################################################################
#   method doNew {}--
#    Dummy method Actual method is in DB code
# Arguments
#   NONE
# 
# Results
#   NONE
# 
  method doAbort {} {
    variable State
    parray State
  }
  
################################################################
#   method doClose {}--
#    Destroy window and associated object
# Arguments
#   NONE
# 
# Results
#   Window, State and object are deleted
# 
  method doClose {} {
    variable State
# puts "CLOSE DESTROY"
    $State(top) destroy
    destroy $State(top)
  }

#
# Search screens
#

################################################################
#   method doSearch {}--
#    Open a search window
# Arguments
#   NONE
# 
# Results
#   New toplevel window is created to search current table
# 
  method doSearch {} {
    variable State

    catch {destroy .tsearch}
    set t [toplevel .tsearch]
    wm title $t "Search Database"
    
    foreach in [array names State label,*] {
      lassign [split $in ,] f fld
      if {![info exists State(combo,$fld)]} {
        lappend flds $State($in)
      }
    }
    ttk::combobox $t.c1 -textvar [my varname State(search,field)] \
        -values [lsort -dict $flds]
    entry $t.e1 -textvar [my varname State(search,value)] 
    label $t.l1 -text "Search for"
    label $t.l2 -text "in"
    grid $t.l1 $t.e1 $t.l2 $t.c1 -row 1
    
    set f [frame $t.bf]
    grid $t.bf -columnspan 4 -row 8
    foreach txt {All Find Advanced Close} \
            hlp {{Find All Matches} {Perform this search} {More search options} {Close this window}} {
      set w [button $f.b_$txt -text $txt -command [list [self] search_$txt]]
      pack $w -side left
      balloon $w $hlp
    }
  }
  
################################################################
#   method search_Close {}--
#    Implement close button in Search window.
# Arguments
#   NONE
# 
# Results
#   Destroys search window
# 
  method search_Close {} {
    destroy .tsearch
  }
  
################################################################
#   method search_Advanced {}--
#    Advanced search window.
# Arguments
#   NONE
# 
# Results
#   New window created
# 
  method search_Advanced {} {
    variable State

    catch {destroy .tsearch.f2}
    catch {destroy .tsearch.bf.b_Advanced}
    set f [frame .tsearch.f2]
    grid $f -row 3 -columnspan 4

    foreach in [array names State label,*] {
      lassign [split $in ,] l fld
      if {![info exists State(combo,$fld)]} {
        lappend flds $State($in)
      }
    }
    ttk::combobox $f.c1 -textvar [my varname State(adv_search,field)] \
        -values [lsort -dict $flds]
    entry $f.e1 -textvar [my varname State(adv_search,value)] 
    label $f.l1 -text "But not "
    label $f.l2 -text "in"
    grid $f.l1 $f.e1 $f.l2 $f.c1 -row 1
    
  }

################################################################
#   method search_All {}--
#    Implements "All" button in search window
# Arguments
#   NONE
# 
# Results
#   Invokes search_Find to modify State array
# 
  method search_All {} {
    variable State
    set State(search,value) %
    set in [lindex [array names State label,*]  0]
    set State(search,field) $State($in)
    my search_Find
  }

################################################################
#   method search_Find {}--
#    Search table for rows that match criteria.
#    Implements "Find" button in search window
# Arguments
#   NONE
# 
# Results
#   Modifies State variable
# 
  method search_Find {} {
    variable State
    if {[string first "%" $State(search,value)] >= 0} {
      set op "like"
    } else {
      set op "="
    }
    
    if {($State(search,field) eq "") ||  ($State(search,value) eq "")} {
      tk_messageBox -type ok -message "Need values in both entry boxes.  \nUse '%' for wildcard"
      return
    }

    set sql "$State(field,$State(search,field)) $op '$State(search,value)'"

    if {[info exists State(adv_search,field)] && ($State(adv_search,field) ne "")} {

      if {($State(adv_search,field) eq "") ||  ($State(adv_search,value) eq "")} {
        tk_messageBox -type ok \
	    -message "Need values in both entry boxes.  \nUse '%' for wildcard"
        return
      }

      if {[string first "%" $State(adv_search,value)] >= 0} {
        set op "NOT LIKE"
      } else {
        set op "!="
      }

      set sql "$sql AND $State(field,$State(adv_search,field)) $op '$State(adv_search,value)'"
    }
    # puts "SQL: $sql"
    my populateBySearch $sql 
    if {[llength $State(rows)] > 1} {
      if {![winfo exists .tsearch.sc]} {
        scale .tsearch.sc -command "[self] populateRow" -from 0 \
            -to [expr {[llength $State(rows)]-1}] \
            -orient horizontal
        grid .tsearch.sc -columnspan 4 -row 10
      } else {
        .tsearch.sc configure -to [expr {[llength $State(rows)]-1}]
      }
    }
  }
  
################################################################
#   method populateRow {pos}--
#    Fill fields in the GUI from a list of potential rows
# Arguments
#   pos		Position in the list to display
# 
# Results
#   GUI and State array is modified
# 
  method populateRow {pos} {
    variable State
    my populate [lindex $State(rows) $pos]
  }

}

if {[info exists argv] && ([lsearch $argv -testET] >= 0)} {
  source helpBalloon.tcl
  
  lassign $argv testED val
  
  set obj [editTable .t1 "" "" -table xyz]
  set obj2 [.t1 clone .t2]
  set obj3 [.t1 clone .t3]
  puts $obj2

if {$val == 1} { 
  $obj makeGUI {
    table create xyz (
      id integer unique,
      loginid text -- loginID --- {User Name} {LoginID} {} {-width 12 -bg green} {1 1}
      fname text -- First name --- {First Name} {Given} {} {-width 12 -bg pink} {2 1}
      lname text -- Last name --- {Last Name} {Surname} {} {-width 16 -bg yellow} {2 2}
      street text -- Address --- {Street} {Addr1} {} {-width 26 -bg yellow} {3 1}
      city text -- Address --- {CityHelp} {City} {} {-width 12 -bg yellow} {4 1}
      state text -- Address --- {StateHelp} {State} {} {-width 26 -bg yellow} 
    );
  }
  grid .t1

  $obj2 makeGUI {
    table create xyz (
      id integer unique,
      phone text, -- foo --- {Help Phone} {Phone} {}
      type text -- bar --- {Help Type} {Type}
    );
  }
  grid .t2
}
if {$val == 2} {
  # Simple, no instrumentation
  $obj3 makeGUI {
    table create phoneList (
      id integer unique,
      name text,  -- person
      phone text, -- phone number
      type text -- home, work, mobile
    );
  }
  grid $obj3
}
if {$val == 3} {
  $obj3 makeGUI {
    table create phoneList (
      id integer unique,
      name text,  --- {First/Last Names} {Name: Last, First} {} {-width 40} {1 1} {-columnspan 2}
      phone text, --- {Number with area code} {Phone Number} {} {} {2 1}
      type text   --- {Home, Work, Cell} {Type} {} {} {2 2}
    );
  }
  grid $obj3
}  

if {$val == 4} {
  oo::objdefine $obj3 method getValues {args} {return [list Work Home Cell]}
  $obj3 makeGUI {
    table create phoneList (
      id integer unique,
      name text,  --- {First/Last Names} {Name: Last, First} {} {-width 40} {1 1} {-columnspan 2}
      num  text, --- {Number with area code} {Phone Number} {} {} {2 1}
      typeid integer references types --- {Home, Work, Cell} {Type} {types.name} {} {2 2}
    );
  }
  grid $obj3
}

if {$val == 5} {
  oo::objdefine $obj3 method getValues {args} {return [list Work Home Cell]}
  oo::objdefine $obj3 method getSchema {args} {return {table create phone {
      id integer unique,
      num  text, 
      personid integer references person,
      typeid integer references types 
    }
  }
  }
  $obj3 config validate,home_phone "$obj3 validatePhone"
  $obj3 makeGUI {table create person (
      id integer unique,
      name text,     --- {First/Last Names} {Name: Last, First} {} {-width 40} {1 1} {-columnspan 3}
      address text,  --- {Address} {Street Address} {} {-width 40} {2 1} {-columnspan 3}
      home_phone string,    --- {Work Phone} {Home Phone} {phone.num} {-width 8} {3 1} {}
      work_phone string,    --- {Work Phone} {Work Phone} {phone.num} {-width 8} {3 2} {}
      cell_phone string,    --- {Work Phone} {Cell Phone} {phone.num} {-width 8} {3 3} {}
    );
  }
  grid $obj3
}

if {$val == 6} {
$obj3 makeGUI {
 table create phoneList (
   id integer unique,
   name text,  --- {First Last} {Name} {} {-width 40} {1 1} {-columnspan 2} 
   num  text, --- {Number with area code} {Phone Number} {} {} {2 1}   
   type text  --- {Type of phone} {Type} {} {} {2 2}
 );
 } 

  grid .t3
  $obj populate {loginid llid fname First lname Last street "123 Any"}
  $obj2 populate {phone 123-456-7890 type Work}
}

# genDict tests
if {$val == 7} {
proc prtLst {lst} {
  foreach l $lst {puts $l}
}
  set sch1 {
    table create foo (
      id int primary key,
      name text,
      address text
    )
  }
  puts "SIMPLE - $sch1"
  prtLst [$obj genDictFromSchema  $sch1]

  set sch2 {
    table create foo (
      id int primary key,  -- Primary key
      name text, --- "NameHelp" "Name Label" {} {-width 10} {1 1} {-columnspan 2}
      address text --- "AddrHelp" "Addr Label" {} {-width 30} {1 2} {-columnspan 2}
    )
  }
  puts "\nInstrumented -- $sch2"
  prtLst [$obj genDictFromSchema  $sch2]
  
}
if {$val == 8} {
  set dicts {
    {fld useID help "Use ID?" label "Use ID" values {yes no} ref {} args {} gridOpts {}}
    {primary id fld useRedact help "Use Redact?" label "Use Redact" values {yes no} ref {} args {} gridOpts {}}
  }
  $obj makeGUIwDict $dicts full
  $obj3 makeGUIwDict $dicts unful
  grid $obj
  grid $obj3
}
if {$val == 9} {
  set sch {
CREATE TABLE subject (
  subjectid string,               --- {Randomized subject ID} {Final Subject ID} {} {-width 10} {1 0}
  nick string,                     --- {Nickname} {Nickname} {} {-width 20} {1 2} {-columnspan 2}
  first_name string,               --- {First Name} {First Name} {study} {-width 20} {2 0}
  middle_name string,              --- {Middle Name} {Middle Name} {study} {-width 20} {2 1}
  last_name string,                --- {Last Name} {Last Name} {study} {-width 20} {2 2}
  );
  }

  puts "DICT: "
  foreach d [$obj  genDictFromSchema $sch] {
    puts $d
  }
  oo::objdefine $obj3 method getValues {args} {puts "GV -- $args --" ; return [list Work Home Cell]}
  $obj3 makeGUI $sch
  grid $obj3

}
if {$val == 10} {
  set sch { TABLE CREATE user (
  id integer unique primary key,
  first_name string,           --- {First Name} {First Name} {} {-width 20} {1 0}
  middle_name string,          --- {Middle Name} {Middle Name} {} {-width 20} {1 1}
  last_name string,            --- {Last Name} {Last Name}  {} {-width 20} {1 2}
  city  string,                --- {City} {City} {} {-width 20} {6 0}
  state string,                --- {State} {State/Province} {} {-width 4} {6 1}
  language string,             --- {Language} {Language} {VALUES {} Italian Russian German {US English} Hindi {Portuguese (Brazilian)} {Spanish (Latin American)} {French Canadian}} {-width 8} {6 2}
  work_email string,           --- {Work Email} {Work Email} {useremail.address} {} {4 0}
  personal_email string,       --- {Personal Email} {Personal Email} {useremail.address} {} {4 1}
  work_phone string,               --- {Work Phone} {Work Phone} {userphone.number} {} {5 0}
  home_phone string,               --- {Home Phone} {Home Phone} {userphone.number} {} {5 1}
  privilege integer references priv,   --- {Privilege Level} {Privileges} {priv.name} {} {2 0}
  userid string,               --- {Unique within site} {User ID} {} {-width 20} {2 1}
  login_name string,           --- {Login Name} {Login Name} {} {-width 20} {3 0}
  password string                  --- {Don't forget this password} {Password} {} {-width 20 -show *} {3 1} 
  )}
  puts "\nO-SHIT\n"
  dict for {k v} [$obj  genDictFromSchema $sch] {
    puts "K: $k V: $v"
  }

}

}
