#####################################################################
# This code is written and copyrighted by Clif Flynt, 2017
#
# It is licenced for public use with these terms:
#
#  This licence agreement must remain intact with the code.
# 
#  You may use this code freely for commercial or non-commercial purposes.
#
#  You may examine and modify the code to your heart's content.
#
#  You may not claim to have authored the code (aside from modifications
#  you may have made.)
#
#  You may not distribute the code without this license agreement
#
#  Maintenance, extension, modification, etc can be performed by:
#
#   Noumena Corporation
#   8888 Black Pine Ln
#   Whitmore Lake, MI  48189
#
#   Contact: clif@noucorp.com
#
#####################################################################

catch {TDBC_support destroy}

 oo::class create TDBC_support {
  
################################################################
#   method init {dbEngine args}--
#    Initialize connection to DB
# Arguments
#   dbEngine	Type of engine 
#   args	Arguments the engine requires
# 
# Results
#   Opens a DB connection
# 
  method init {dbEngine args} {
    variable State
    set State(db) [::tdbc::${dbEngine}::connection create db {*}$args]
  }
  
################################################################
#   destructor
#    Destroy and close.
# Arguments
#   NONE
# 
# Results
#   DB connection is closed
# 
  destructor {
    variable State
    $State(db) close
  }

################################################################
#   method getValueByRef {tbl fld primary value}--
#    Return the value in a field based on primary key and value
# Arguments
#   tbl		Name of table
#   fld		Name of field to return a value for
#   primary	Name of primary field 
#   value	Value for primary.
# 
# Results
#   No side effects
# 
  method getValueByRef {tbl fld primary value} {
    variable State
    set rtn [$State(db) allrows  "SELECT $fld FROM $tbl WHERE $primary = :value"]
    return [lindex $rtn 0 1]
  }
  
################################################################
#   method populateBySearch {search}--
#    Return values from a search
# Arguments
#   search	SQL test phrase
# 
# Results
#   Update State(rows)
# 
  method populateBySearch {search} {
    variable State
    set State(rows) [$State(db) allrows "SELECT * FROM $State(table) WHERE $search"]
    my populate [lindex $State(rows) 0]
  }

  
################################################################
#   method getSchema {tbl} --
#    Return the schema for a table
# Arguments
#   tbl		Name of the table to get Schema for
# Results
#   Modifies State(table) and State(primary) if save==1
# 
  method getSchema {tbl} {
    variable State
    set State(table) $tbl
    set dd {*}[$State(db) primarykeys $tbl]
    set prim [dict get $dd columnName]
    set State(primary) $prim
    return [dict get [dict get [$State(db) tables] $tbl] sql]
  }
  
################################################################
#   method doSQL {sql}--
#    Evaluate an SQL string
# Arguments
#   sql		complete SQL command
# 
# Results
#   Returns results of SQL cmd
# 
  method doSQL {sql} {
    variable State
    set x [$State(db) prepare $sql]
    $x execute
  }

################################################################
#   method getValues {refList}--
#    Return a list of 
# Arguments
#   refList	List of fields for a reference
#    reference list would be things like tbl.field1 tbl.field2 that
#    are referenced by some primary key to tbl
# 
# Results
#   Modifies State
# 
  method getValues {refList} {
    variable State
    set rtn ""
    
    foreach ref $refList {
      lassign [split $ref .] tbl fld
    
      set dd {*}[$State(db) primarykeys $tbl]
      set prim [dict get $dd columnName]
      set State(primary,$ref) $prim

      set o [$State(db) prepare "SELECT $prim, $fld FROM $tbl"]
      
      set pos 0
      foreach s [$o allrows -as lists] {
        lassign $s k v
        lappend State(ids,$ref) $k
        lset rtn $pos "[lindex $rtn $pos] $v"
	incr pos
      }
    }
    return $rtn
  }

################################################################
#   method doReplace {}--
#    Replace currently selected row with edited values
# Arguments
#   NONE
# 
# Results
#   Database is modified.
# 
  method doReplace {} {
    variable State
    set cmd {}
    foreach in [array names State value,*] {
      lassign [split $in ,] v fld
      if {[info exists State(combo,$fld)]} {
        set $fld [lindex $State(ids,$State(ref,$fld)) [$State(combo,$fld) current]]
      } else {
        set $fld $State($in)
      }
      if {$fld ne $State(primary)} {
        append cmd "$fld = :$fld, "
      }
    }
    
    set cmd [string trim $cmd " ,"]
    
    set sql "UPDATE $State(table) SET $cmd WHERE $State(primary) = [set $State(primary)]"
    
    # puts "SQL: $sql"
    $State(db) allrows $sql
  }

################################################################
#   method doNew {}--
#    Add currently displayed values as a new row
# Arguments
#   NONE
# 
# Results
#   DB is modified
# 
  method doNew {} {
    variable State
    set cmd {}
    foreach in [array names State value,*] {
      lassign [split $in ,] v fld
      if {[info exists State(combo,$fld)]} {
        set $fld [lindex $State(ids,$State(ref,$fld)) [$State(combo,$fld) current]]
      } else {
        set $fld $State($in)
      }

      if {![info exists State(primary)] || ($fld ne $State(primary))} {
        lappend flds $fld
	lappend vals :$fld
      }
    }
    
    set cmd [string trim $cmd " ,"]
    
    set sql "INSERT INTO $State(table) ([join $flds ,]) VALUES ([join $vals ,])"
    
    # puts "SQL: $sql"
    $State(db) allrows $sql
    
    set new [$State(db) allrows "SELECT max($State(primary)) FROM $State(table)"]
    set State(value,$State(primary)) [lindex $new 0 1]
  }

################################################################
#   method closeDB {}--
#    Close the connection
# Arguments
#   NONE
# 
# Results
#   DB connection is closed
# 
  method closeDB {} {
    variable State
    $State(db) close
  }

}

if {[info exists argv] && ([lsearch $argv -testET_TDBC] >= 0)} {

catch {editTable destroy}
catch {destroy .t1}

source editTablesOO.tcl

source helpBalloon.tcl
puts [package require tdbc::sqlite3]
file delete test.db

set obj [editTable .t1 TDBC_support sqlite3 -dbArgs test.db]
set db [$obj config db]

$db allrows {
  CREATE TABLE person (
    id integer unique primary key,
    loginid text, -- loginID --- {User Name} {LoginID} {} {-width 12 -bg green} {1 1}
    fname text, -- First name --- {First Name} {Given} {} {-width 12 -bg pink} {2 1}
    lname text, -- Last name --- {Last Name} {Surname} {} {-width 16 -bg yellow} {2 2}
    addrRef integer references addr -- Address Ref --- {} {Street Address} {addr.street} {-width 24} {3 1}
    );
    CREATE TABLE addr (
    id integer unique primary key,
    street text, -- Address --- {Street} {Addr1} {} {-width 26 -bg yellow} {3 1}
    city text, -- Address --- {CityHelp} {City} {} {-width 12 -bg yellow} {4 1}
    state text -- Address --- {StateHelp} {State} {} {-width 26 -bg yellow} 
  )
}

$db allrows {INSERT INTO addr VALUES (1, '123 St', 'Acity', 'AA')}
$db allrows {INSERT INTO addr VALUES (2, '234 St', 'Bcity', 'BB')}
$db allrows {INSERT INTO person VALUES (1, 'aaa', 'Alpha', 'Adam', 1)}
$db allrows {INSERT INTO person VALUES (2, 'bbb', 'Beta', 'Blocker', 2)}
$db allrows {INSERT INTO person VALUES (3, 'aa2', 'Abel', 'Adam', 2)}
$db allrows {INSERT INTO person VALUES (4, 'aa3', 'Aard', 'Adam', 2)}

$obj makeGUI [$obj getSchema person]
$obj config -table person
$obj populateBySearch "loginid = 'aaa'"

pack .t1
}
